/*!
 * @edoc/resolver-qa 
 * (c) 2020-2022 Alex
 * Released under the MIT License.
 */
'use strict';

var share = require('@esydoc/share');
var chalk = require('chalk');
var glob = require('glob');
var path = require('path');
var fs = require('fs-extra');
var tapable = require('tapable');
var pathFilter = require('@esydoc/path-filter');
var docletParser = require('@esydoc/doclet-parser');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var glob__default = /*#__PURE__*/_interopDefaultLegacy(glob);
var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var runtimeApiMap = {
    getFn: 'getFn',
    createExcutorWrapperFn: 'createExcutorWrapperFn',
    tryTransfromDeps: 'tryTransfromDeps',
    shouldExistTaseCase: 'shouldExistTaseCase'
};

var TestCaseNode = /** @class */ (function (_super) {
    __extends(TestCaseNode, _super);
    function TestCaseNode(options) {
        var _this = _super.call(this) || this;
        _this.options = options;
        _this.args = new share.GetValueNode({ perperty: options.args });
        _this.expect = new share.GetValueNode({ perperty: options.expect });
        return _this;
    }
    // done -> next
    TestCaseNode.prototype.genDoneFnByUserHack = function (doneKey) {
        var onDone = this.options.onDone;
        if (!onDone)
            return doneKey;
        return "\n    function doneByUserControl(e, ret) {\n      " + onDone + "(e, ret, " + doneKey + ")\n    }\n    ";
    };
    TestCaseNode.prototype.generate = function () {
        var _a = this.options, apiPath = _a.apiPath, categroySort = _a.categroySort, _b = _a.deps, deps = _b === void 0 ? [] : _b;
        var shouldInjectDeps = (this.args.isFn || this.expect.isFn) && deps.length > 0;
        return "\n      it('{" + apiPath + "} - \u7B2C" + categroySort + "\u4E2A\u6D4B\u8BD5\u7528\u4F8B', (done) => {\n        const testRunner = apiMark['" + apiPath + "'].value\n        const doneCb = " + this.genDoneFnByUserHack('done') + "\n        \n        " + (
        // 解析依赖
        shouldInjectDeps
            ? "const deps=" + runtimeApiMap.tryTransfromDeps + "(testRunner.deps, testRunner)"
            : '') + "\n\n        const args = " + this.args.generate(shouldInjectDeps ? 'deps' : '') + "\n        const expectValue = " + this.expect.generate(shouldInjectDeps ? 'deps' : '') + "\n        testRunner(args, expectValue, doneCb)\n      })\n    ";
    };
    return TestCaseNode;
}(share.CodeNode));

var createNode = function (value) {
    return {
        value: value,
        nextSibling: null,
        child: null,
        return: null
    };
};
var toBTree = function (rawTree) {
    var root = createNode(rawTree.value);
    var current;
    var last;
    if (rawTree.deps) {
        rawTree.deps.forEach(function (subValue, i) {
            if (i === 0) {
                current = last = root.child = createNode(subValue);
                current.return = root;
            }
            else {
                current = createNode(subValue);
                last.nextSibling = current;
                current.return = last;
                last = current;
            }
        });
    }
    return root;
};
// 因为我们要做依赖分析，所以优先兄弟节点遍历
function beginWork(root, cb) {
    var stack = [];
    var current = root;
    while (current || stack.length) {
        if (typeof current !== 'undefined') {
            if (cb(current))
                return current;
            current.child && stack.push(current.child);
            current.nextSibling && stack.push(current.nextSibling); // 先child后nextSibling，这样做是因为我们要先访问兄弟节点
        }
        current = stack.pop();
    }
    return null;
}
var traverseParentPathCheckCircularRef = function (node) {
    var val = node.value;
    var parent = node.return;
    while (parent !== null) {
        // 节点往上遍历，检查是否有循环依赖
        if (parent.value === val) {
            throw new Error("The value " + val + " be circulared reference by parent path");
        }
        parent = parent.return;
    }
};
var combineNode = function (node, hostRoot) {
    if (hostRoot.child !== null) {
        if (node.child === null) {
            node.child = hostRoot.child;
            hostRoot.child.return = node;
            beginWork(node.child, function (internalNode) {
                // 检查循环依赖
                traverseParentPathCheckCircularRef(internalNode);
            });
        }
    }
    else if (hostRoot.nextSibling !== null) {
        // hostRoot压根就没有nextSibling，这段代码为了以后
        if (node.nextSibling === null) {
            node.nextSibling = hostRoot.nextSibling;
            hostRoot.nextSibling.return = node;
            beginWork(node.nextSibling, function (internalNode) {
                // 检查循环依赖
                traverseParentPathCheckCircularRef(internalNode);
            });
        }
    }
};
var combineForest2BTree = function (forest) {
    var rootTree = forest.shift();
    (function forest2tree() {
        if (!forest.length)
            return;
        var hostRoot = forest.shift();
        if (hostRoot && rootTree) {
            var matchNode = beginWork(rootTree, function (node) {
                return node.value === hostRoot.value;
            });
            if (matchNode !== null) {
                combineNode(matchNode, hostRoot);
            }
            else {
                // rootTree增加一个兄弟节点
                var current = rootTree;
                while (current !== null) {
                    if (current.nextSibling !== null) {
                        current = current.nextSibling;
                    }
                    else {
                        current.nextSibling = hostRoot;
                        hostRoot.return = current;
                        break;
                    }
                }
            }
        }
        forest2tree();
    })();
    return rootTree;
};
function shrink(path) {
    var temp = {};
    path.forEach(function (val) {
        temp[val] = 1;
    });
    return Object.keys(temp);
}
function analysisDeps(rawForest) {
    if (rawForest.length === 0)
        return [];
    var depPath = [];
    var bForest = rawForest.map(toBTree);
    var fullTree = combineForest2BTree(bForest);
    if (fullTree) {
        beginWork(fullTree, function (node) {
            depPath.push(node.value);
        });
    }
    return shrink(depPath.reverse());
}

function outputTestCaseCode(options) {
    var testCaseNode = new TestCaseNode(options);
    return runtimeApiMap.shouldExistTaseCase + "('" + options.apiPath + "', apiQueue) && " + testCaseNode.generate();
}
var TestCaseNum = 0;
var ApiQACodeGenerator = /** @class */ (function (_super) {
    __extends(ApiQACodeGenerator, _super);
    function ApiQACodeGenerator(options) {
        var _this = _super.call(this) || this;
        _this.nodeConfigMap = new Map();
        _this.hooks = {
            genRuntimeCode: new tapable.SyncWaterfallHook(['code'])
        };
        _this.options = options;
        return _this;
    }
    ApiQACodeGenerator.prototype.collectNode = function (config) {
        this.nodeConfigMap.set(config.node.path, config);
    };
    ApiQACodeGenerator.prototype.ouputNewNodeConfigQueueByDeps = function () {
        console.log(chalk__default['default'].magenta('Begining analyze Dependencies...'));
        var rawForest = [];
        var nodeConfigMap = this.nodeConfigMap;
        nodeConfigMap.forEach(function (config) {
            var node = config.node;
            rawForest.push({
                value: config.node.path,
                deps: node.deps
            });
        });
        var flatenTree = analysisDeps(rawForest);
        var sortConigQueue = [];
        flatenTree.forEach(function (path) {
            var config = nodeConfigMap.get(path);
            config && sortConigQueue.push(config);
        });
        return sortConigQueue;
    };
    ApiQACodeGenerator.prototype.getHead = function () {
        var markItems = [];
        this.nodeConfigMap.forEach(function (config) {
            markItems.push(new share.MarkItemNode(config.node.path, config.node.generate()));
        });
        return "\n      export const apiMark = " + share.MarkItemNode.genMarkObjectCode(markItems) + "\n      export const hostContext = " + share.GetValueNode.toString(this.options.hostContext || {}) + "\n    ";
    };
    ApiQACodeGenerator.prototype.getBody = function () {
        // gen test case by ApiCallNode
        return this.ouputNewNodeConfigQueueByDeps()
            .map(function (config) {
            var node = config.node, expectQueue = config.expectQueue, args2d = config.args2d, onDone = config.onDone;
            // 兼容一维数组
            if (typeof args2d === 'function' || // args2d is () => {}
                !args2d.length || // args2d is []
                (args2d.find(function (item) { return !Array.isArray(item); }) &&
                    !Array.isArray(expectQueue)) // args2d is [{}, number, string, ...]
            ) {
                return [
                    outputTestCaseCode({
                        apiPath: node.path,
                        args: args2d,
                        expect: expectQueue[0],
                        sort: ++TestCaseNum,
                        categroySort: 1,
                        deps: node.deps,
                        onDone: onDone
                    })
                ];
            }
            else {
                if (Array.isArray(args2d) && Array.isArray(expectQueue)) {
                    // 二维数组
                    return args2d.map(function (args, j) {
                        return outputTestCaseCode({
                            apiPath: node.path,
                            args: args,
                            expect: expectQueue[j],
                            sort: ++TestCaseNum,
                            categroySort: j + 1,
                            deps: node.deps,
                            onDone: onDone
                        });
                    });
                }
                else {
                    throw new Error("The args & expect must be an array.");
                }
            }
        })
            .reduce(function (acc, item) {
            acc = acc.concat(item);
            return acc;
        }, [])
            .join('\n');
    };
    ApiQACodeGenerator.prototype.getFoot = function () {
        var staicPath = path__default['default'].resolve(__dirname, '../static');
        var filePaths = glob__default['default']
            .sync('*.js', { cwd: staicPath })
            .map(function (file) { return path__default['default'].join(staicPath, file); });
        var content = filePaths.reduce(function (code, filePath) {
            code += '\n' + fs__default['default'].readFileSync(filePath).toString();
            return code;
        }, '');
        return this.hooks.genRuntimeCode.call(content); // 方便业务方注入业务逻辑
    };
    ApiQACodeGenerator.prototype.generate = function () {
        if (this.nodeConfigMap.size === 0)
            return '';
        return "\n     import sinon from 'sinon'\n\n      " + this.getHead() + "\n\n      export function injectTestSuite (apiQueue) {\n        describe('QA testing', () => {\n          " + this.getBody() + "\n        })\n      }\n\n      " + this.getFoot() + "\n    ";
    };
    return ApiQACodeGenerator;
}(share.CodeGenerator));

var hash = 0;
var ApiCallNode = /** @class */ (function (_super) {
    __extends(ApiCallNode, _super);
    function ApiCallNode(options) {
        var _this = _super.call(this) || this;
        _this.assert = 'helper.assert'; // 断言函数，以后可配置
        _this.id = hash++;
        _this.path = options.path;
        _this.pathPrefix = options.pathPrefix;
        _this.deps = options.deps || [];
        return _this;
    }
    Object.defineProperty(ApiCallNode.prototype, "fullPath", {
        get: function () {
            return this.pathPrefix + "." + this.path;
        },
        enumerable: false,
        configurable: true
    });
    ApiCallNode.prototype.generate = function () {
        return "\n      // created by ApiCallNode\n      (function () {\n        const fn =  " + runtimeApiMap.getFn + "(() => " + this.fullPath + ")\n        const spyFn = sinon.spy(fn)\n        return " + runtimeApiMap.createExcutorWrapperFn + "(spyFn, '" + this.path + "', " + JSON.stringify(this.deps) + ", " + this.id + ", " + this.assert + ")\n      })()\n    ";
    };
    return ApiCallNode;
}(share.CodeNode));

var templateMap = {
    'hyext-qa-miniapp': path__default['default'].resolve(__dirname, '../templates/qa-miniapp')
};
var ApiQACodeResolver = /** @class */ (function (_super) {
    __extends(ApiQACodeResolver, _super);
    function ApiQACodeResolver(options) {
        var _this = _super.call(this) || this;
        _this.templateMap = templateMap;
        _this.hooks = {
            config: new tapable.SyncWaterfallHook(['config'])
        };
        _this.options = Object.assign({
            pathPrefix: '',
            verbose: true
        }, options);
        _this.generator = new ApiQACodeGenerator({
            hostContext: _this.options.output.hostContext
        });
        _this.check = pathFilter.createCheckPathFn(options);
        var plugins = options.plugins;
        if (plugins && plugins.length) {
            plugins.forEach(function (plugin) {
                plugin.apply(_this);
            });
        }
        return _this;
    }
    ApiQACodeResolver.prototype._selectApiConfig = function (apiConfig) {
        return Object.assign({
            args: [],
            deps: [],
            onDone: undefined
        }, apiConfig ? apiConfig.qa : {});
    };
    ApiQACodeResolver.prototype.apply = function (scheduler) {
        var _this = this;
        var pathPrefix = this.options.pathPrefix;
        this.scheduler = scheduler;
        scheduler.hooks.schedule.tap('ApiQACodeResolver', function (data) {
            var ast = data.ast, config = data.config;
            var _a = _this.hooks.config.call(config), source = _a.source, apiConfig = _a.apiConfig;
            if (source.configPath && !_this.check(source.configPath))
                return;
            var qa = _this._selectApiConfig(apiConfig);
            var apiCallNode = new ApiCallNode(__assign({ deps: qa.deps, pathPrefix: pathPrefix }, source));
            _this.generator.collectNode({
                node: apiCallNode,
                args2d: qa.args,
                expectQueue: [formatRoot(ast === null || ast === void 0 ? void 0 : ast.ret)],
                onDone: qa.onDone
            });
        });
        scheduler.hooks.emit.tapPromise('ApiQACodeResolver', function (results) { return __awaiter(_this, void 0, void 0, function () {
            var helper, code, key, beautifulCode;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        helper = this.scheduler.helper;
                        code = this.generator.generate();
                        key = this.key;
                        if (!(code === '')) return [3 /*break*/, 1];
                        results.push({
                            status: 'fail',
                            msg: '解析代码为空',
                            key: key
                        });
                        return [3 /*break*/, 3];
                    case 1:
                        beautifulCode = helper.perttyCode(code);
                        return [4 /*yield*/, helper.addFile(path__default['default'].join(templateMap['hyext-qa-miniapp'], 'qa-test-runner.js'), beautifulCode)];
                    case 2:
                        _a.sent();
                        results.push({
                            status: 'success',
                            key: key
                        });
                        _a.label = 3;
                    case 3: return [2 /*return*/];
                }
            });
        }); });
        scheduler.hooks.collect.tap('ApiQACodeResolver', function (config) {
            var update = config.update, apiConfig = config.apiConfig;
            if (!apiConfig.qa) {
                apiConfig.qa = _this._selectApiConfig(null);
                update = true;
            }
            return { update: update, apiConfig: apiConfig };
        });
    };
    return ApiQACodeResolver;
}(share.EventResolver));
function formatRoot(tree) {
    if (tree == null)
        return tree;
    tree = docletParser.clean(docletParser.unwrap(tree));
    // To mark tree is data tree after tree sharking.
    tree.id = docletParser.DATA_NODE_ID_PREFIX + 'root';
    return tree;
}

module.exports = ApiQACodeResolver;
