import { UI } from '@hyext/hy-ui'
import React, { Component } from 'react'
import { TouchableOpacity } from 'react-native'
import StatusButton from '../../components/StatusButton'
import { hostContext } from '../../qa-test-runner'

import {
  API_QA_STATUS,
  API_QA_STATUS_ICON,
  API_QA_STATUS_ICON_COLOR,
  RUN_TIME_STATUS,
  RUN_TIME_STATUS_TEXT,
  RUN_TIME_STATUS_BTN_TEXT
} from '../../constants'

import './index.hycss'

const {
  View,
  Text,
  ScrollView,
  TopviewGetInstance,
  Tab,
  Dialog,
  Input,
  Tip,
  Form,
  Switch
} = UI

let testCount = 0

class QAPanel extends Component {
  lastTestStats = [0, 0, 0]

  $tab = null

  constructor() {
    super()

    this.state = {
      type: 'all'
    }
  }

  getCurrentTypeList(type) {
    const { rawApiList } = this.props

    let list
    if (type === 'all') {
      list = [...rawApiList]
    } else {
      list = rawApiList.filter(v => v.scope === type)
    }

    return list
  }

  handlePress(apiInfo) {
    this.props.history.push('/tests', apiInfo)
  }

  renderApiList() {
    return this.getCurrentTypeList(this.state.type)
      .filter(v => Boolean(v))
      .map((api, i) => {
        return (
          <StatusButton
            iconType={API_QA_STATUS_ICON[api.status]}
            iconColor={API_QA_STATUS_ICON_COLOR[api.status]}
            text={api.path}
            onPress={() => {
              this.handlePress(api, i)
            }}
            key={i}
          ></StatusButton>
        )
      })
  }

  calcCount() {
    let pass = 0
    let fail = 0
    let skip = 0

    this.getCurrentTypeList(this.state.type).forEach(apiInfo => {
      if (apiInfo) {
        switch (apiInfo.status) {
          case API_QA_STATUS.PASS:
            ++pass
            break
          case API_QA_STATUS.FAIL:
            ++fail
            break
          case API_QA_STATUS.SKIP:
            ++skip
            break
        }
      }
    })

    return (this.lastTestStats = [pass, fail, skip])
  }

  reset(cb) {
    if (this.onReset) {
      this.onReset()
    }

    cb()
  }

  startTest = () => {
    const { onStartTest } = this.props
    this.reset(() => {
      testCount++
      const newList = this.getCurrentTypeList(this.state.type)
      onStartTest && onStartTest(newList)
    })
  }

  startReport = () => {
    if (testCount === 0) {
      Tip.show('兄弟莫着急，先跑测试')
      return
    }

    const { type } = this.state
    const { onStartReport } = this.props
    const [successNum, failNum, skipNum] = this.lastTestStats

    onStartReport &&
      onStartReport({
        scope: type,
        apiList: this.getCurrentTypeList(type),
        successNum,
        failNum,
        skipNum,
        projectName: hostContext.projectName || '默认',
        template: 1
      })
  }

  handleConfirm = () => {
    const { onStartReport } = this.props
    const { commitMsg, type } = this.state
    const [successNum, failNum, skipNum] = this.lastTestStats

    if (commitMsg == '') {
      Tip.show('请先填写备注')

      setTimeout(() => {
        this.$dialog.open()
      }, 500)

      return
    }

    onStartReport &&
      onStartReport({
        comment: commitMsg,
        scope: type,
        apiList: this.getCurrentTypeList(type),
        successNum,
        failNum,
        skipNum,
        projectName: hostContext.projectName || '默认',
        template: 1,
        host: global.__os
      })
  }

  getTabList() {
    const rawList = this.props.rawApiList
    const map = {}

    rawList.forEach(item => {
      map[item.scope] = 1
    })

    const extra = Object.keys(map).map(item => {
      return {
        value: item,
        label: item
      }
    })

    return [
      {
        value: 'all',
        label: 'all'
      },
      ...extra
    ]
  }

  handleTabChange = (item, index) => {
    this.$tab.scrollTo(index)
    this.setState({ type: item.value })
  }

  render() {
    const totalCount = this.calcCount()

    const runtimeStatus = this.props.runtimeStatus
    const runtimeText = RUN_TIME_STATUS_TEXT[runtimeStatus]
    const disableStart = runtimeStatus === RUN_TIME_STATUS.ING
    const runtimeBtnText = RUN_TIME_STATUS_BTN_TEXT[runtimeStatus]
    const tapList = this.getTabList()

    return (
      <>
        <View className="title">
          <Text className="title-text">
            {hostContext.projectName || 'QA小程序'}
          </Text>
        </View>
        <Tab
          ref={c => {
            this.$tab = c
          }}
          data={tapList}
          scrollable
          value={this.state.type}
          onChange={this.handleTabChange}
          dataItemContainerStyle={{ height: 40 }}
        ></Tab>
        <ScrollView className="scroller">
          <View class="api-list">{this.renderApiList()}</View>
        </ScrollView>
        <TotalMemo
          skipCount={totalCount[2]}
          failCount={totalCount[1]}
          successCount={totalCount[0]}
          runtime={runtimeText}
        ></TotalMemo>
        <FloatButton
          text={runtimeBtnText}
          disabled={disableStart}
          onPress={this.startTest}
        ></FloatButton>
        <FloatButton
          text={'上报'}
          disabled={disableStart}
          onPress={this.startReport}
          className="float-btn-report"
          color="rgb(242, 50, 68)"
        ></FloatButton>
      </>
    )
  }
}

const TotalMemo = React.memo(function Total({
  skipCount,
  failCount,
  successCount,
  runtime
}) {
  return (
    <View className="total">
      <TotalItem
        name="成功数"
        color={'rgb(97, 203, 40)'}
        num={successCount}
      ></TotalItem>
      <TotalItem
        name="失败数"
        color={'rgb(242, 50, 68)'}
        num={failCount}
      ></TotalItem>
      <TotalItem
        name="跳过数"
        color={'rgb(24, 138, 250)'}
        num={skipCount}
      ></TotalItem>
      <TotalItem name="状态:" color={'#aaa'} num={runtime}></TotalItem>
    </View>
  )
})

function TotalItem({ name, num, color }) {
  return (
    <View className="total-item">
      <Text className="total-item-name">{name}</Text>
      <Text className="total-item-num" style={{ color }}>
        {num}
      </Text>
    </View>
  )
}

class FloatButton extends Component {
  static defaultProps = {
    disabled: false,
    onPress: null,
    text: '',
    style: null,
    color: 'rgb(255, 132, 0)'
  }

  id = null

  componentDidUpdate() {
    TopviewGetInstance().replace(this.renderButton(), this.id)
  }

  renderButton() {
    const { disabled, onPress, text, style, color } = this.props
    return (
      <TouchableOpacity
        disabled={disabled}
        onPress={onPress}
        className={`float-btn ${disabled ? 'float-btn-disable' : ''}`}
        style={[{ borderColor: color }, style]}
      >
        <Text style={{ color }}>{text}</Text>
      </TouchableOpacity>
    )
  }

  componentDidMount() {
    setTimeout(() => {
      TopviewGetInstance()
        .add(this.renderButton())
        .then(id => {
          this.id = id
        })
    })
  }

  componentWillUnmount() {
    if (this.id !== null) {
      TopviewGetInstance().remove(this.id)
    }
  }

  render() {
    return null
  }
}

export default QAPanel
